
\predicates {
  in64(int);
  inU64(int);

  in32(int);
  inU32(int);

  in16(int);
  inU16(int);

  in8(int);
  inU8(int);

  inU1(int);
}

\functions {
// Signed 64-bit arithmetic
  \partial \relational int add64(int, int);
  \partial \relational int sub64(int, int);
  \partial \relational int mul64(int, int);
  \partial \relational int div64(int, int);
  \partial \relational int minus64(int);
  \partial \relational int shift64(int, int);
  \partial \relational int bitneg64(int);

// Unsigned 64-bit arithmetic
  \partial \relational int addU64(int, int);
  \partial \relational int subU64(int, int);
  \partial \relational int mulU64(int, int);
  \partial \relational int divU64(int, int);
  \partial \relational int minusU64(int);
  \partial \relational int shiftU64(int, int);
  \partial \relational int bitnegU64(int);

// Signed 32-bit arithmetic
  \partial \relational int add32(int, int);
  \partial \relational int sub32(int, int);
  \partial \relational int mul32(int, int);
  \partial \relational int div32(int, int);
  \partial \relational int minus32(int);
  \partial \relational int shift32(int, int);
  \partial \relational int bitneg32(int);

// Unsigned 32-bit arithmetic
  \partial \relational int addU32(int, int);
  \partial \relational int subU32(int, int);
  \partial \relational int mulU32(int, int);
  \partial \relational int divU32(int, int);
  \partial \relational int minusU32(int);
  \partial \relational int shiftU32(int, int);
  \partial \relational int bitnegU32(int);

// Signed 16-bit arithmetic
  \partial \relational int add16(int, int);
  \partial \relational int sub16(int, int);
  \partial \relational int mul16(int, int);
  \partial \relational int div16(int, int);
  \partial \relational int minus16(int);
  \partial \relational int shift16(int, int);
  \partial \relational int bitneg16(int);

// Unsigned 16-bit arithmetic
  \partial \relational int addU16(int, int);
  \partial \relational int subU16(int, int);
  \partial \relational int mulU16(int, int);
  \partial \relational int divU16(int, int);
  \partial \relational int minusU16(int);
  \partial \relational int shiftU16(int, int);
  \partial \relational int bitnegU16(int);

// Signed 8-bit arithmetic
  \partial \relational int add8(int, int);
  \partial \relational int sub8(int, int);
  \partial \relational int mul8(int, int);
  \partial \relational int div8(int, int);
  \partial \relational int minus8(int);
  \partial \relational int shift8(int, int);
  \partial \relational int bitneg8(int);

// Unsigned 8-bit arithmetic
  \partial \relational int addU8(int, int);
  \partial \relational int subU8(int, int);
  \partial \relational int mulU8(int, int);
  \partial \relational int divU8(int, int);
  \partial \relational int minusU8(int);
  \partial \relational int shiftU8(int, int);
  \partial \relational int bitnegU8(int);

// (Unsigned) 1-bit arithmetic
  \partial \relational int addU1(int, int);
  \partial \relational int bitnegU1(int);

// Casts to the particular integer types (from any other type)
  \partial \relational int castU32(int);
  \partial \relational int cast32(int);

  \partial \relational int castU16(int);
  \partial \relational int cast16(int);

  \partial \relational int castU8(int);
  \partial \relational int cast8(int);

// Bit-wise boolean operations that are independent of
// the number of bits
  \partial \relational int and(int, int);
  \partial \relational int andBit(int, int);
  \partial \relational int or(int, int);

  \partial \relational int shiftLeft(int, int);
  \partial \relational int shiftRight(int, int);

  \partial \relational int ones(int);
  \partial \relational int zeros(int);

// General (unbounded) multiplication
  \partial \relational int mul(int, int);

// Modulo, which we assume not reveal any overflows (correct?)
  \partial \relational int mod(int, int);

// General (unbounded) integer division
  \partial \relational int div(int, int);

// Reflected equality
  \partial \relational int equals(int, int);
}

\problem {

////////////////////////////////////////////////////////////////////////////////
// Signed 64-bit arithmetic

  \forall int x; (in64(x) -> x >= -8*1024*1024*1024*1024*1024*1024 &
                             x < 8*1024*1024*1024*1024*1024*1024)
&
// Commented, because this will in general cause nontermination
//  \forall int x; (x >= -8*1024*1024*1024*1024*1024*1024 &
//                    x < 8*1024*1024*1024*1024*1024*1024 -> in64(x))
// &

// Signed 64-bit addition

  \forall int x, y, res; {add64(x, y)} (
    add64(x, y) = res ->
    \exists int k; res = x + y + k * 16*1024*1024*1024*1024*1024*1024 &
    in64(res)
  )
&

// Signed 64-bit subtraction

  \forall int x, y; {sub64(x, y)} sub64(x, y) = add64(x, -y)
&

// Signed 64-bit multiplication

  \forall int x, y, res; {mul64(x, y)} (
    mul64(x, y) = res ->
    \exists int k; res = mul(x, y) + k * 16*1024*1024*1024*1024*1024*1024 &
    in64(res)
  )
&

// Signed 64-bit division

  \forall int x, y, res; {div64(x, y)} (div64(x, y) = res -> (
    x = -8*1024*1024*1024*1024*1024*1024 & y = -1 &
    res = -8*1024*1024*1024*1024*1024*1024
    |
    !(x = -8*1024*1024*1024*1024*1024*1024 & y = -1) &
    res = div(x, y)
  ))
&

// Signed 64-bit negation

  \forall int x; {minus64(x)} minus64(x) = add64(0, -x)
&

// Bit-shifts to the left or right

  \forall int x, y, res; {shift64(x, y)}
      (y >= 0 -> shift64(x, y) = res ->
       \exists int k;
         res = shiftLeft(x, y) + k * 16*1024*1024*1024*1024*1024*1024 &
       in64(res))
&
  \forall int x, y, res; {shift64(x, y)}
      (y < 0 -> shift64(x, y) = shiftRight(x, -y))
&

// Bit-wise negation

  \forall int x; {bitneg64(x)} bitneg64(x) = -x - 1
&


////////////////////////////////////////////////////////////////////////////////
// Unsigned 64-bit arithmetic

  \forall int x; (inU64(x) -> x >= 0 & x < 16*1024*1024*1024*1024*1024*1024)
&

// Unsigned 64-bit addition

  \forall int x, y, res; {addU64(x, y)} (
    addU64(x, y) = res ->
    \exists int k; res = x + y + k * 16*1024*1024*1024*1024*1024*1024 &
    inU64(res)
  )
&

// Unsigned 64-bit subtraction

  \forall int x, y; {subU64(x, y)} subU64(x, y) = addU64(x, -y)
&

// Unsigned 64-bit multiplication

  \forall int x, y, res; {mulU64(x, y)} (
    mulU64(x, y) = res ->
    \exists int k; res = mul(x, y) + k * 16*1024*1024*1024*1024*1024*1024 &
    inU64(res)
  )
&

// Unsigned 64-bit division

  \forall int x, y; {divU64(x, y)} divU64(x, y) = div(x, y)
&

// Unsigned 64-bit negation

  \forall int x; {minusU64(x)} minusU64(x) = addU64(0, -x)
&

// Bit-shifts to the left or right

  \forall int x, y, res; {shiftU64(x, y)}
      (y >= 0 -> shiftU64(x, y) = res ->
       \exists int k;
         res = shiftLeft(x, y) + k * 16*1024*1024*1024*1024*1024*1024 &
       inU64(res))
&
  \forall int x, y, res; {shiftU64(x, y)}
      (y < 0 -> shiftU64(x, y) = shiftRight(x, -y))
&

// Bit-wise negation

  \forall int x; {bitnegU64(x)} bitnegU64(x) =
                  16*1024*1024*1024*1024*1024*1024 - x - 1
&


////////////////////////////////////////////////////////////////////////////////
// Signed 32-bit arithmetic

  \forall int x; (in32(x) -> x >= -2*1024*1024*1024 & x < 2*1024*1024*1024)
&
// Commented, because this will in general cause nontermination
//  \forall int x; (x >= -2*1024*1024*1024 & x < 2*1024*1024*1024 -> in32(x))
//&

// Signed 32-bit addition

  \forall int x, y, res; {add32(x, y)} (
    add32(x, y) = res ->
    \exists int k; res = x + y + k * 4*1024*1024*1024 &
    in32(res)
  )
&

// Signed 32-bit subtraction

  \forall int x, y; {sub32(x, y)} sub32(x, y) = add32(x, -y)
&

// Signed 32-bit multiplication

  \forall int x, y, res; {mul32(x, y)} (
    mul32(x, y) = res ->
    \exists int k; res = mul(x, y) + k * 4*1024*1024*1024 &
    in32(res)
  )
&

// Signed 32-bit division

  \forall int x, y, res; {div32(x, y)} (div32(x, y) = res -> (
    x = -2*1024*1024*1024 & y = -1 &
    res = -2*1024*1024*1024
    |
    !(x = -2*1024*1024*1024 & y = -1) &
    res = div(x, y)
  ))
&

// Signed 32-bit negation

  \forall int x; {minus32(x)} minus32(x) = add32(0, -x)
&

// Bit-shifts to the left or right

  \forall int x, y, res; {shift32(x, y)}
      (y >= 0 -> shift32(x, y) = res ->
       \exists int k;
         res = shiftLeft(x, y) + k * 4*1024*1024*1024 &
       in32(res))
&
  \forall int x, y, res; {shift32(x, y)}
      (y < 0 -> shift32(x, y) = shiftRight(x, -y))
&

// Bit-wise negation

  \forall int x; {bitneg32(x)} bitneg32(x) = -x - 1
&


////////////////////////////////////////////////////////////////////////////////
// Unsigned 32-bit arithmetic

  \forall int x; (inU32(x) -> x >= 0 & x < 4*1024*1024*1024)
&

// Unsigned 32-bit addition

  \forall int x, y, res; {addU32(x, y)} (
    addU32(x, y) = res ->
    \exists int k; res = x + y + k * 4*1024*1024*1024 &
    inU32(res)
  )
&

// Unsigned 32-bit subtraction

  \forall int x, y; {subU32(x, y)} subU32(x, y) = addU32(x, -y)
&

// Unsigned 32-bit multiplication

  \forall int x, y, res; {mulU32(x, y)} (
    mulU32(x, y) = res ->
    \exists int k; res = mul(x, y) + k * 4*1024*1024*1024 &
    inU32(res)
  )
&

// Unsigned 32-bit division

  \forall int x, y; {divU32(x, y)} divU32(x, y) = div(x, y)
&

// Unsigned 32-bit negation

  \forall int x; {minusU32(x)} minusU32(x) = addU32(0, -x)
&

// Bit-shifts to the left or right

  \forall int x, y, res; {shiftU32(x, y)}
      (y >= 0 -> shiftU32(x, y) = res ->
       \exists int k;
         res = shiftLeft(x, y) + k * 4*1024*1024*1024 &
       inU32(res))
&
  \forall int x, y, res; {shiftU32(x, y)}
      (y < 0 -> shiftU32(x, y) = shiftRight(x, -y))
&

// Bit-wise negation

  \forall int x; {bitnegU32(x)} bitnegU32(x) = 4*1024*1024*1024 - x - 1
&


////////////////////////////////////////////////////////////////////////////////
// Signed 16-bit arithmetic

  \forall int x; (in16(x) -> x >= -32*1024 & x < 32*1024)
&
// Commented, because this will in general cause nontermination
//  \forall int x; (x >= -32*1024 & x < 32*1024 -> in16(x))
//&

// Signed 16-bit addition

  \forall int x, y, res; {add16(x, y)} (
    add16(x, y) = res ->
    \exists int k; res = x + y + k * 64*1024 &
    in16(res)
  )
&

// Signed 16-bit subtraction

  \forall int x, y; {sub16(x, y)} sub16(x, y) = add16(x, -y)
&

// Signed 16-bit multiplication

  \forall int x, y, res; {mul16(x, y)} (
    mul16(x, y) = res ->
    \exists int k; res = mul(x, y) + k * 64*1024 &
    in16(res)
  )
&

// Signed 16-bit division

  \forall int x, y, res; {div16(x, y)} (div16(x, y) = res -> (
    x = -32*1024 & y = -1 &
    res = -32*1024
    |
    !(x = -32*1024 & y = -1) &
    res = div(x, y)
  ))
&

// Signed 16-bit negation

  \forall int x; {minus16(x)} minus16(x) = add16(0, -x)
&

// Bit-shifts to the left or right

  \forall int x, y, res; {shift16(x, y)}
      (y >= 0 -> shift16(x, y) = res ->
       \exists int k;
         res = shiftLeft(x, y) + k * 64*1024 &
       in16(res))
&
  \forall int x, y, res; {shift16(x, y)}
      (y < 0 -> shift16(x, y) = shiftRight(x, -y))
&

// Bit-wise negation

  \forall int x; {bitneg16(x)} bitneg16(x) = -x - 1
&


////////////////////////////////////////////////////////////////////////////////
// Unsigned 16-bit arithmetic

  \forall int x; (inU16(x) -> x >= 0 & x < 64*1024)
&

// Unsigned 16-bit addition

  \forall int x, y, res; {addU16(x, y)} (
    addU16(x, y) = res ->
    \exists int k; res = x + y + k * 64*1024 &
    inU16(res)
  )
&

// Unsigned 16-bit subtraction

  \forall int x, y; {subU16(x, y)} subU16(x, y) = addU16(x, -y)
&

// Unsigned 16-bit multiplication

  \forall int x, y, res; {mulU16(x, y)} (
    mulU16(x, y) = res ->
    \exists int k; res = mul(x, y) + k * 64*1024 &
    inU16(res)
  )
&

// Unsigned 16-bit division

  \forall int x, y; {divU16(x, y)} divU16(x, y) = div(x, y)
&

// Unsigned 16-bit negation

  \forall int x; {minusU16(x)} minusU16(x) = addU16(0, -x)
&

// Bit-shifts to the left or right

  \forall int x, y, res; {shiftU16(x, y)}
      (y >= 0 -> shiftU16(x, y) = res ->
       \exists int k;
         res = shiftLeft(x, y) + k * 64*1024 &
       inU16(res))
&
  \forall int x, y, res; {shiftU16(x, y)}
      (y < 0 -> shiftU16(x, y) = shiftRight(x, -y))
&

// Bit-wise negation

  \forall int x; {bitnegU16(x)} bitnegU16(x) = 64*1024 - x - 1
&


////////////////////////////////////////////////////////////////////////////////
// Signed 8-bit arithmetic

  \forall int x; (in8(x) -> x >= -128 & x < 128)
&
// Commented, because this will in general cause nontermination
//  \forall int x; (x >= -128 & x < 128 -> in8(x))
//&

// Signed 8-bit addition

  \forall int x, y, res; {add8(x, y)} (
    add8(x, y) = res ->
    \exists int k; res = x + y + k * 256 &
    in8(res)
  )
&

// Signed 8-bit subtraction

  \forall int x, y; {sub8(x, y)} sub8(x, y) = add8(x, -y)
&

// Signed 8-bit multiplication

  \forall int x, y, res; {mul8(x, y)} (
    mul8(x, y) = res ->
    \exists int k; res = mul(x, y) + k * 256 &
    in8(res)
  )
&

// Signed 8-bit division

  \forall int x, y, res; {div8(x, y)} (div8(x, y) = res -> (
    x = -128 & y = -1 &
    res = -128
    |
    !(x = -128 & y = -1) &
    res = div(x, y)
  ))
&

// Signed 8-bit negation

  \forall int x; {minus8(x)} minus8(x) = add8(0, -x)
&

// Bit-shifts to the left or right

  \forall int x, y, res; {shift8(x, y)}
      (y >= 0 -> shift8(x, y) = res ->
       \exists int k;
         res = shiftLeft(x, y) + k * 256 &
       in8(res))
&
  \forall int x, y, res; {shift8(x, y)}
      (y < 0 -> shift8(x, y) = shiftRight(x, -y))
&

// Bit-wise negation

  \forall int x; {bitneg8(x)} bitneg8(x) = -x - 1
&

////////////////////////////////////////////////////////////////////////////////
// Unsigned 8-bit arithmetic

  \forall int x; (inU8(x) -> x >= 0 & x < 256)
&

// Unsigned 8-bit addition

  \forall int x, y, res; {addU8(x, y)} (
    addU8(x, y) = res ->
    \exists int k; res = x + y + k * 256 &
    inU8(res)
  )
&

// Unsigned 8-bit subtraction

  \forall int x, y; {subU8(x, y)} subU8(x, y) = addU8(x, -y)
&

// Unsigned 8-bit multiplication

  \forall int x, y, res; {mulU8(x, y)} (
    mulU8(x, y) = res ->
    \exists int k; res = mul(x, y) + k * 256 &
    inU8(res)
  )
&

// Unsigned 8-bit division

  \forall int x, y; {divU8(x, y)} divU8(x, y) = div(x, y)
&

// Unsigned 8-bit negation

  \forall int x; {minusU8(x)} minusU8(x) = addU8(0, -x)
&

// Bit-shifts to the left or right

  \forall int x, y, res; {shiftU8(x, y)}
      (y >= 0 -> shiftU8(x, y) = res ->
       \exists int k;
         res = shiftLeft(x, y) + k * 256 &
       inU8(res))
&
  \forall int x, y, res; {shiftU8(x, y)}
      (y < 0 -> shiftU8(x, y) = shiftRight(x, -y))
&

// Bit-wise negation

  \forall int x; {bitnegU8(x)} bitnegU8(x) = 256 - x - 1
&


////////////////////////////////////////////////////////////////////////////////
// (Unsigned) 1-bit arithmetic

  \forall int x; (inU1(x) -> x >= 0 & x < 2)
&

// Unsigned 1-bit addition

  \forall int x, y, res; {addU1(x, y)} (
    addU1(x, y) = res ->
    \exists int k; res = x + y + k * 2 &
    inU1(res)
  )
&

// Bit-wise negation

  \forall int x; {bitnegU1(x)} bitnegU1(x) = 2 - x - 1
&

////////////////////////////////////////////////////////////////////////////////
// Casts between signed and unsigned arithmetic

  \forall int x, res; {castU32(x)} (
    castU32(x) = res ->
    \exists int k; res = x + k * 4*1024*1024*1024 &
    inU32(res)
  )
&
  \forall int x, res; {cast32(x)} (
    cast32(x) = res ->
    \exists int k; res = x + k * 4*1024*1024*1024 &
    in32(res)
  )
&

  \forall int x, res; {castU16(x)} (
    castU16(x) = res ->
    \exists int k; res = x + k * 64*1024 &
    inU16(res)
  )
&
  \forall int x, res; {cast16(x)} (
    cast16(x) = res ->
    \exists int k; res = x + k * 64*1024 &
    in16(res)
  )
&

  \forall int x, res; {castU8(x)} (
    castU8(x) = res ->
    \exists int k; res = x + k * 256 &
    inU8(res)
  )
&
  \forall int x, res; {cast8(x)} (
    cast8(x) = res ->
    \exists int k; res = x + k * 256 &
    in8(res)
  )
&


////////////////////////////////////////////////////////////////////////////////

  \forall int x, y, res; {shiftLeft(x, y)} (
    shiftLeft(x, y) = res -> (
    y > 0 & res = shiftLeft(2*x, y-1)
    |
    y <= 0 & res = x
  ))
&

  \forall int x, y, res; {shiftRight(x, y)} (
    shiftRight(x, y) = res ->
    \exists int subres, diff; (
      subres = shiftLeft(res, y) & diff = shiftLeft(1, y) & (
      x >= 0 & subres <= x & subres + diff > x
      |
      x < 0 & subres >= x & subres - diff < x
    )))
&

// Bit-wise and. This mainly does a case analysis over the second
// argument, which means that it is usually more
// efficient to have constant values as the second argument

  \forall int x; {and(x, 0)} and(x, 0) = 0
&
  \forall int x; {and(x, -1)} and(x, -1) = x
&
  \forall int x; {andBit(x, 0)} andBit(x, 0) = 0
&
  \forall int x; {andBit(x, -1)} andBit(x, -1) = x
&
  \forall int x, y, res; {andBit(x, y)}
      ((y >= 1 | y <= -2) -> andBit(x, y) = res ->
       \exists int k, l, m, n, subres; (
           andBit(k, l) = subres &
           x = 2*k + m & y = 2*l + n & m >= 0 & m <= 1 & (
             n = 0 & res = subres * 2
             |
             n = 1 & res = subres * 2 + m
       )))
&

// Optimised version for common cases (y is of the form 2^n-1)
  \forall int x, y, res; {and(x, y)}
      ((y >= 1 | y <= -2) -> and(x, y) = res ->
       \exists int n1, n2, threshold; (
           n1 = ones(y) & threshold = shiftLeft(1, n1)
           &
           (
             n1 > 0 &
             y >= 0 & y < threshold &
                // HACK to prevent the "mod" from escaping
                // (should be fixed in the function encoder, TODO)
             \exists int x'; (x' = x & res = mod(x', threshold))
             |
             !(n1 > 0 & y >= 0 & y < threshold) &
             \exists int x'; (x' = x & res = andBit(x', y))
       )))
&

  \forall int x, res; {ones(x)} (ones(x) = res ->
    x = 0 & res = 0
    |
    x = -1 & res = -1
    |
    (x > 0 | x < -1) &
    \exists int k; (
      x = 2*k & res = 0
      |
      x = 2*k + 1 & res = 1 + ones(k)
  ))
&
  \forall int x; {zeros(x)} zeros(x) = ones(-x-1)
&

// Bit-wise or

  \forall int x, y; {or(x, y)} or(x, y) = -and(-x-1, -y-1) - 1
&

////////////////////////////////////////////////////////////////////////////////

// General multiplication (on the unbounded integers)

  \forall int x; {mul(x, 0)} mul(x, 0) = 0
&
  \forall int x; {mul(x, -1)} mul(x, -1) = -x
&
  \forall int x, y, res; {mul(x, y)}
      ((y >= 1 | y <= -2) -> mul(x, y) = res ->
       \exists int l, n, subres; (
           mul(2*x, l) = subres &
           y = 2*l + n & (
             n = 0 & res = subres
             |
             n = 1 & res = subres + x
       )))
&

////////////////////////////////////////////////////////////////////////////////

// Modulo

  \forall int x, y, res; {mod(x, y)}
      (mod(x, y) = res ->
       \exists int k; mul(k, y) + res = x &
       0 <= res & (res < y | res < -y))
&  

// Division

  \forall int x, y, res; {div(x, y)}
      \exists int mulres; (
         mul(div(x, y), y) = mulres &
         mulres <= x & (mulres > x + y | mulres > x - y)
      )
&  

////////////////////////////////////////////////////////////////////////////////

// Reflected equality

  \forall int x; {equals(x, x)} equals(x, x) = 1
&
  \forall int x, y; {equals(x, y)} (x != y -> equals(x, y) = 0)



// everything is negated (the definitions are premisses)
->
  false
}
